import sys
from typing import Any

from ..._core import Dispatcher
from ..._events import CleanupEvent, ScenarioFailEvent, ScenarioPassEvent
from ..plugin import Plugin


class Terminator(Plugin):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self._passed = 0
        self._failed = 0

    def subscribe(self, dispatcher: Dispatcher) -> None:
        dispatcher.listen(ScenarioPassEvent, self.on_scenario_pass) \
                  .listen(ScenarioFailEvent, self.on_scenario_fail) \
                  .listen(CleanupEvent, self.on_cleanup, priority=-1)

    def on_scenario_pass(self, event: ScenarioPassEvent) -> None:
        self._passed += 1

    def on_scenario_fail(self, event: ScenarioFailEvent) -> None:
        self._failed += 1

    def on_cleanup(self, event: CleanupEvent) -> None:
        if self._failed > 0 or self._passed == 0:
            sys.exit(1)
        else:
            sys.exit(0)



from ...._core import Dispatcher
from .._reporter import Reporter


class SilentReporter(Reporter):
    def __init__(self, verbosity: int) -> None:
        self._verbosity = verbosity

    def subscribe(self, dispatcher: Dispatcher) -> None:
        pass



from typing import Any, Set, Type

__all__ = ("Event",)


class EventRegistry:
    events: Set[str] = set()

    @classmethod
    def register(cls, event: Type['Event']) -> None:
        cls.events.add(event.__name__)

    @classmethod
    def is_registered(cls, event: Type['Event']) -> bool:
        return event.__name__ in cls.events


class Event:
    def __init_subclass__(cls, **kwargs: Any) -> None:
        if EventRegistry.is_registered(cls):
            raise Exception(f"Event {cls} already registered")
        EventRegistry.register(cls)



from ._dispatcher import Dispatcher
from ._runner import Runner
from ._virtual_scenario import ScenarioStatus, VirtualScenario
from ._virtual_step import StepStatus, VirtualStep

__all__ = ("Dispatcher", "Runner", "VirtualScenario", "ScenarioStatus",
           "VirtualStep", "StepStatus",)



from .director import Director
from .plugin import Plugin
from .skipper import Skipper
from .terminator import Terminator
from .validator import Validator

__all__ = ("Director", "Plugin", "Skipper", "Terminator", "Validator",)



import inspect
from types import MethodType
from typing import Any, Union


class StepStatus:
    PASSED = "PASSED"
    FAILED = "FAILED"


class VirtualStep:
    def __init__(self, method: MethodType) -> None:
        self._method: MethodType = method
        self._status: Union[str, None] = None

    @property
    def name(self) -> str:
        return self._method.__name__

    def is_coro(self) -> bool:
        return inspect.iscoroutinefunction(self._method)

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        return self._method(*args, **kwargs)

    def mark_passed(self) -> None:
        self._status = StepStatus.PASSED

    def is_passed(self) -> bool:
        return self._status == StepStatus.PASSED

    def mark_failed(self) -> None:
        self._status = StepStatus.FAILED

    def is_failed(self) -> bool:
        return self._status == StepStatus.FAILED



import inspect
from typing import Any, Callable, Dict, List, Tuple, Type

from .._events import Event

__all__ = ("Dispatcher",)


HandlerType = Callable[..., Any]


class Dispatcher:
    def __init__(self) -> None:
        self._events: Dict[str, List[Tuple[int, HandlerType]]] = {}

    def register(self, subscriber: Any) -> None:  # subscriber: Plugin
        subscriber.subscribe(self)

    def listen(self, event: Type[Event], handler: HandlerType, priority: int = 0) -> "Dispatcher":
        if event.__name__ not in self._events:
            self._events[event.__name__] = []
        self._events[event.__name__].append((priority, handler))
        return self

    async def fire(self, event: Event) -> None:
        if event.__class__.__name__ not in self._events:
            return
        registered = self._events[event.__class__.__name__]
        for priority, handler in sorted(registered, key=lambda x: x[0], reverse=True):
            if inspect.iscoroutinefunction(handler):
                await handler(event)
            else:
                handler(event)



import asyncio
import signal
import sys
from types import FrameType
from typing import Optional

from ._context import context
from ._core import Runner
from ._interface import Interface
from ._params import params
from ._scenario import Scenario
from ._version import version
from .plugins.skipper import only, skip
from .plugins.validator import Validator

__version__ = version
__all__ = ("Scenario", "Interface", "Runner", "run", "only", "skip", "params", "context",)


def run(*, validator: Optional[Validator] = None) -> None:
    runner = Runner(validator=validator)
    event = asyncio.Event()

    def signal_handler(sig: int, frame: FrameType) -> None:
        event.set()
        sys.exit(1)
    signal.signal(signal.SIGINT, signal_handler)
    asyncio.run(runner.run(event))



import inspect
import os
from typing import Any, Generator, List, Type, Union

from .._scenario import Scenario
from ._virtual_step import VirtualStep


class ScenarioStatus:
    PASSED = "PASSED"
    FAILED = "FAILED"
    SKIPPED = "SKIPPED"


class VirtualScenario:
    def __init__(self, scenario: Type[Scenario]) -> None:
        self._scenario: Type[Scenario] = scenario
        self._status: Union[str, None] = None
        self._path: str = inspect.getfile(scenario)
        self._steps: List[VirtualStep] = []
        self._errors: List[str] = []
        self._exception: Any = None
        self._scope: Union[Scenario, None] = None
        for step in self._scenario.__dict__:
            if not step.startswith("_"):
                method = getattr(self._scenario, step)
                if inspect.isfunction(method):
                    self._steps.append(VirtualStep(method))

    @property
    def subject(self) -> Any:
        return getattr(self._scenario, "subject", None)

    @property
    def path(self) -> str:
        return self._path

    @property
    def namespace(self) -> str:
        return os.path.dirname(os.path.relpath(self._path, "scenarios"))

    @property
    def filename(self) -> str:
        basename = os.path.basename(self._path)
        filename, _ = os.path.splitext(basename)
        return filename

    @property
    def errors(self) -> List[str]:
        return self._errors

    @property
    def exception(self) -> Any:
        return self._exception

    @property
    def scenario(self) -> Type[Scenario]:
        return self._scenario

    @property
    def scope(self) -> Union[Scenario, None]:
        return self._scope

    def set_scope(self, scope: Scenario) -> None:
        self._scope = scope

    def set_errors(self, errors: List[Any]) -> None:
        self._errors = errors

    def set_exception(self, exception: Any) -> None:
        self._exception = exception

    def get_steps(self) -> Generator[VirtualStep, None, None]:
        for step in self._steps:
            yield step

    def __call__(self) -> Any:
        return self._scenario()

    def mark_passed(self) -> None:
        self._status = ScenarioStatus.PASSED

    def is_passed(self) -> bool:
        return self._status == ScenarioStatus.PASSED

    def mark_failed(self) -> None:
        self._status = ScenarioStatus.FAILED

    def is_failed(self) -> bool:
        return self._status == ScenarioStatus.FAILED

    def mark_skipped(self) -> None:
        self._status = ScenarioStatus.SKIPPED

    def is_skipped(self) -> bool:
        return self._status == ScenarioStatus.SKIPPED



import importlib
import os
import sys
from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser
from asyncio import Event
from inspect import isclass
from pathlib import Path
from typing import Any, AsyncGenerator, List, Optional, Tuple, Type

from .._events import (
    ArgParsedEvent,
    ArgParseEvent,
    CleanupEvent,
    ScenarioFailEvent,
    ScenarioPassEvent,
    ScenarioRunEvent,
    ScenarioSkipEvent,
    StartupEvent,
    StepFailEvent,
    StepPassEvent,
)
from .._scenario import Scenario
from ..plugins import Director, Skipper, Terminator, Validator
from ._dispatcher import Dispatcher
from ._virtual_scenario import VirtualScenario
from ._virtual_step import VirtualStep


class Runner:
    def __init__(self, *, validator: Optional[Validator] = None) -> None:
        self._validator: Validator = validator if validator else Validator()

    def load(self, path: str) -> List[Type[Scenario]]:
        name = os.path.splitext(path)[0].replace("/", ".")
        module = importlib.import_module(name)

        scenarios = []
        for name in dir(module):
            val = getattr(module, name)
            if isclass(val) and val.__name__.startswith("Scenario"):
                assert issubclass(val, Scenario), f"{val} must be a subclass of vedro.Scenario"
                scenarios.append(val)
        return scenarios

    def discover(self, root: str) -> List[VirtualScenario]:
        start_dir = os.path.relpath(root)

        scenarios = []
        for path, _, files in os.walk(start_dir):
            for filename in files:
                if not filename.endswith(".py"):
                    continue
                if filename.startswith(".") or filename.startswith("_"):
                    continue
                module_scenarios = self.load(os.path.join(path, filename))
                for scenario in module_scenarios:
                    scenarios.append(VirtualScenario(scenario))

        def cmp(scn: VirtualScenario) -> Tuple[Any, ...]:
            path = Path(scn.path)
            return (len(path.parts),) + path.parts
        scenarios.sort(key=cmp)
        return scenarios

    async def _run_steps(self, scenario: VirtualScenario) -> AsyncGenerator[VirtualStep, None]:
        scope = scenario()
        scenario.set_scope(scope)
        for step in scenario.get_steps():
            try:
                if step.is_coro():
                    await step(scope)
                else:
                    step(scope)
            except:  # noqa: E722
                exception = sys.exc_info()
                scenario.set_exception(exception)
                step.mark_failed()
                yield step
            else:
                step.mark_passed()
                yield step

    async def run(self, event: Event) -> None:
        os.chdir(os.path.dirname(os.path.join(os.getcwd(), sys.argv[0])))

        arg_parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter)

        dispatcher = Dispatcher()
        dispatcher.register(self._validator)
        dispatcher.register(Director())
        dispatcher.register(Skipper())
        dispatcher.register(Terminator())

        await dispatcher.fire(ArgParseEvent(arg_parser))
        args = arg_parser.parse_args()
        await dispatcher.fire(ArgParsedEvent(args))

        scenarios = self.discover("scenarios")
        await dispatcher.fire(StartupEvent(scenarios))

        for scenario in scenarios:
            if event.is_set():
                break

            if scenario.is_skipped():
                await dispatcher.fire(ScenarioSkipEvent(scenario))
                continue

            await dispatcher.fire(ScenarioRunEvent(scenario))

            async for step in self._run_steps(scenario):
                if step.is_failed():
                    await dispatcher.fire(StepFailEvent(step))
                    scenario.mark_failed()
                    break
                elif step.is_passed():
                    await dispatcher.fire(StepPassEvent(step))
                else:
                    raise ValueError()
            else:
                scenario.mark_passed()

            if scenario.is_passed():
                await dispatcher.fire(ScenarioPassEvent(scenario))
            elif scenario.is_failed():
                await dispatcher.fire(ScenarioFailEvent(scenario))
            else:
                raise ValueError()

        await dispatcher.fire(CleanupEvent())



from argparse import ArgumentParser, Namespace
from typing import List

from .._core._virtual_scenario import VirtualScenario
from .._core._virtual_step import VirtualStep
from ._event import Event


class ArgParseEvent(Event):
    def __init__(self, arg_parser: ArgumentParser) -> None:
        self._arg_parser = arg_parser

    @property
    def arg_parser(self) -> ArgumentParser:
        return self._arg_parser


class ArgParsedEvent(Event):
    def __init__(self, args: Namespace) -> None:
        self._args = args

    @property
    def args(self) -> Namespace:
        return self._args


class StartupEvent(Event):
    def __init__(self, scenarios: List[VirtualScenario]) -> None:
        self._scenarios = scenarios

    @property
    def scenarios(self) -> List[VirtualScenario]:
        return self._scenarios


class ScenarioSkipEvent(Event):
    def __init__(self, scenario: VirtualScenario) -> None:
        self._scenario = scenario

    @property
    def scenario(self) -> VirtualScenario:
        return self._scenario


class ScenarioRunEvent(Event):
    def __init__(self, scenario: VirtualScenario) -> None:
        self._scenario = scenario

    @property
    def scenario(self) -> VirtualScenario:
        return self._scenario


class ScenarioFailEvent(Event):
    def __init__(self, scenario: VirtualScenario) -> None:
        self._scenario = scenario

    @property
    def scenario(self) -> VirtualScenario:
        return self._scenario


class ScenarioPassEvent(Event):
    def __init__(self, scenario: VirtualScenario) -> None:
        self._scenario = scenario

    @property
    def scenario(self) -> VirtualScenario:
        return self._scenario


class StepFailEvent(Event):
    def __init__(self, step: VirtualStep) -> None:
        self._step = step

    @property
    def step(self) -> VirtualStep:
        return self._step


class StepPassEvent(Event):
    def __init__(self, step: VirtualStep) -> None:
        self._step = step

    @property
    def step(self) -> VirtualStep:
        return self._step


class CleanupEvent(Event):
    pass


__all__ = ("Event", "ArgParseEvent", "ArgParsedEvent", "StartupEvent",
           "ScenarioSkipEvent", "ScenarioRunEvent", "ScenarioFailEvent", "ScenarioPassEvent",
           "StepFailEvent", "StepPassEvent", "CleanupEvent",)



from typing import Any, Type, Union

from ..._scenario import Scenario


def only(scenario_or_nothing: Union[Type[Scenario], None] = None) -> Any:
    def wrapped(scenario: Type[Scenario]) -> Type[Scenario]:
        setattr(scenario, "__vedro__only__", True)
        return scenario

    if scenario_or_nothing is None:
        return wrapped
    elif issubclass(scenario_or_nothing, Scenario):
        return wrapped(scenario_or_nothing)
    else:
        raise TypeError()



import json
from time import monotonic
from traceback import format_exception
from typing import Union

from rich.console import Console
from rich.style import Style

from ...._core import Dispatcher
from ...._core._virtual_scenario import VirtualScenario
from ...._events import (
    CleanupEvent,
    ScenarioFailEvent,
    ScenarioPassEvent,
    ScenarioRunEvent,
    ScenarioSkipEvent,
    StartupEvent,
)
from .._reporter import Reporter


class RichReporter(Reporter):
    def __init__(self, verbosity: int) -> None:
        super().__init__()
        self._verbosity = verbosity
        self._console = Console(highlight=False, force_terminal=True)
        self._namespace: Union[str, None] = None
        self._start_time = 0.0
        self._total = 0
        self._passed = 0
        self._failed = 0
        self._skipped = 0

    def subscribe(self, dispatcher: Dispatcher) -> None:
        dispatcher.listen(StartupEvent, self.on_startup) \
                  .listen(ScenarioRunEvent, self.on_scenario_run) \
                  .listen(ScenarioSkipEvent, self.on_scenario_skip) \
                  .listen(ScenarioPassEvent, self.on_scenario_pass) \
                  .listen(ScenarioFailEvent, self.on_scenario_fail) \
                  .listen(CleanupEvent, self.on_cleanup)

    def on_startup(self, event: StartupEvent) -> None:
        self._start_time = monotonic()
        self._console.print("Scenarios")

    def on_scenario_skip(self, event: ScenarioSkipEvent) -> None:
        self._skipped += 1

    def _format_subject(self, scenario: VirtualScenario) -> str:
        subject = scenario.subject
        if subject and isinstance(subject, str):
            scope = scenario.scope
            return subject.format(**scope.__dict__)
        return scenario.filename.replace("_", " ")

    def on_scenario_run(self, event: ScenarioRunEvent) -> None:
        self._total += 1

        namespace = event.scenario.namespace
        namespace = namespace.replace("_", " ").replace("/", " / ")
        if namespace != self._namespace:
            self._namespace = namespace
            self._console.print(f"* {namespace}", style=Style(bold=True))

    def on_scenario_pass(self, event: ScenarioPassEvent) -> None:
        self._passed += 1
        subject = self._format_subject(event.scenario)

        self._console.print(f" ✔ {subject}", style=Style(color="green"))

    def on_scenario_fail(self, event: ScenarioFailEvent) -> None:
        self._failed += 1

        subject = self._format_subject(event.scenario)
        self._console.print(f" ✗ {subject}", style=Style(color="red"))

        if self._verbosity > 0:
            for step in event.scenario.get_steps():
                if step.is_passed():
                    self._console.print(f"    ✔ {step.name}", style=Style(color="green"))
                elif step.is_failed():
                    self._console.print(f"    ✗ {step.name}", style=Style(color="red"))

        if self._verbosity > 1:
            tb = format_exception(*event.scenario.exception)
            self._console.print("".join(tb), style=Style(color="yellow"))
            if event.scenario.errors:
                errors = " - " + "\n - ".join(event.scenario.errors)
                self._console.print(errors, style=Style(color="yellow"))
            self._console.print()

        if self._verbosity > 2:
            scope = event.scenario.scope
            if scope.__dict__:
                self._console.print("Scope:", style=Style(color="blue", bold=True))
                for key, val in scope.__dict__.items():
                    self._console.print(f" {key}: ", end="", style=Style(color="blue"))
                    try:
                        val_repr = json.dumps(val, ensure_ascii=False, indent=4)
                    except Exception:
                        val_repr = repr(val)
                    self._console.print(val_repr)
                self._console.print()

    def on_cleanup(self, event: CleanupEvent) -> None:
        elapsed = monotonic() - self._start_time
        if self._failed == 0 and self._passed > 0:
            style = Style(color="green", bold=True)
        else:
            style = Style(color="red", bold=True)
        self._console.print()
        self._console.print(f"# {self._total} scenarios, "
                            f"{self._failed} failed, "
                            f"{self._skipped} skipped "
                            f"[blue]({elapsed:.2f}s)",
                            style=style)



from typing import Any, Optional

from ..._core import Dispatcher
from ..._events import ScenarioFailEvent, ScenarioRunEvent
from ..plugin import Plugin

__all__ = ("Validator",)


class Validator(Plugin):
    def __init__(self, validator: Optional[Any] = None) -> None:
        self._validator = validator

    def subscribe(self, dispatcher: Dispatcher) -> None:
        dispatcher.listen(ScenarioRunEvent, self.on_scenario_run) \
                  .listen(ScenarioFailEvent, self.on_scenario_fail)

    def on_scenario_run(self, event: ScenarioRunEvent) -> None:
        event.scenario.set_errors([])

    def on_scenario_fail(self, event: ScenarioFailEvent) -> None:
        if self._validator:
            event.scenario.set_errors(event.scenario.errors + self._validator.errors())
            self._validator.reset()



from typing import Any, Union, cast

from ..._core import Dispatcher
from ..._events import ArgParsedEvent, ArgParseEvent
from ..plugin import Plugin
from ._reporter import Reporter
from .reporters import RichReporter, SilentReporter


class Director(Plugin):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self._dispatcher: Union[Dispatcher, None] = None
        self._reporter: Union[Reporter, None] = None
        self._reporters = {
            'rich': RichReporter,
            'silent': SilentReporter,
        }
        self._default_reporter = 'rich'

    def subscribe(self, dispatcher: Dispatcher) -> None:
        dispatcher.listen(ArgParseEvent, self.on_arg_parse) \
                  .listen(ArgParsedEvent, self.on_arg_parsed)
        self._dispatcher = dispatcher

    def on_arg_parse(self, event: ArgParseEvent) -> None:
        event.arg_parser.add_argument("-r", "--reporter",
                                      choices=self._reporters, default=self._default_reporter)
        event.arg_parser.add_argument("-v", "--verbose", action="count", default=0)

    def on_arg_parsed(self, event: ArgParsedEvent) -> None:
        self._reporter = self._reporters[event.args.reporter](event.args.verbose)
        self._reporter.subscribe(cast(Dispatcher, self._dispatcher))
